#!/usr/bin/perl
#
# convert_to_3d_rules.pl
# Wed Apr  5 13:29:00 EDT 2006
# Tue Oct 10 17:01:25 EDT 2006
#
# usage:  convert_to_3d_rules.pl -infile input_file -config config_file -outpath output_dir 
#
# This file reads a 2d rules file and performs the following operations
#
# First, config_file is read.  The following lines may be included in the config_file:
#
#      GLOBAL_PARAM {param_name}
#                 Indicates that {param_name} is a global parameter that applies to 
#                 all tiers.  Non-global parameters will be replicated for each tier
#
#      TIER_OFFSET  {number}
#                 Indicates that GDS layer numbers will be offset by the tier number
#                 times {number}.  Non-specific layers have offset zero.
#
#      CREATE_NONSPECIFIC_LAYERS {0 || 1}
#                 If set to zero, then non-specific layers are generated.  Otherwise
#                 these LAYER lines are omitted.  Default is 1.
#
#      VALID_RANGE_MAX {number}
#                 Include only layers with source GDS numbers less than {number}.  This
#                 must be less than TIER_OFFSET.  If undefined, VALID_RANGE_MAX defaults
#                 to (TIER_OFFSET-1).
#
#      NUMBER_OF_TIERS {number}
#                 Sets the number of tiers for the output rules file.
#      
#      INCLUDE_FILE {filename}
#                 Incidates that layer definitions in {filename} will be used in the 
#                 rules file being processed.
#
#      REM {comment_text}
#                 Indicates a comment line to be ignored
#
#
# If the rules file has the string "master_top" in its filename, then it will be treated 
# as a special global definitions rules file.  Otherwise, normal processing will occur.
#
#
# Processing of "master_top" rules files:
#
# A master_top rules file sets up the configuration of each tier in the process.  It must
# be customized after processing to set applicable parameters for each tier.  When a master_top 
# rules file is processed, the configuration file should include USE_TITLE, USE_FZSUB, 
# RECOMMEND_CHK, DENSITY_MIM_CHK, DENSITY_MAX_CHK, DENSITY_PRT and PRC_CHECK001 as 
# GLOBAL_PARAMs.  The default parameter section will be processed first, with non-global default 
# parameters duplicated for eaach tier.  The default settings will be based on the source file 
# and will likely be incorrect.
#
# When the processing reaches the #IFDEF USE_PROCESS process-specific variable section, all
# process specific preprocessor directive sets will be maintained as is.  
# 
# Comments and SVRF statements such as PRECISION and UNIT LENGTH will be maintained as is.
# 
# LAYER statements will be replicated NUMBER_OF_TIERS times, plus one additional time for
# a nonspecific tier if CREATE_NONSPECIFIC_LAYERS=1.  The prefix T#_, where # is the tier 
# number, will be added to each tier.  Non-specific tiers are given the prefix NS_.  LAYER
# statements in the output file will have layer numbers offset by the tier number times
# TIER_OFFSET, with zero offset for the NS tier.
#
# If the source layer GDS number is greater than VALID_RANGE_MAX, then the LAYER statement
# is commented out.
#
# 
# Processing of other rules files:
#
# For all other rules file inputs, a layer catalog will be prepared that includes all original
# and derived layers.  Then, an output file will be generated that includes the
# following sections.  First, #DEFINE statements that define parameters listed as
# GLOBAL_PARAMs will be identified and a note will be created at the top of the file that
# will instuct the user to manually consolidate the global #DEFINE statements.  This note will
# include a listing of all global #DEFINE statements found in the file.  (Since these may be part
# of nested #IFDEF blocks, these are best edited by hand.) 
#
# Then macro definitions will be relocated to the top of the file.  Then the entire file will be 
# replicated for each tier, not including the NS tier.  The GLOBAL_PARAM #DEFINE  and macro 
# definition lines will be commented out in the tier-specific sections of the output rules file. 
# Include statements are relocated to the top of the file in the same way.
#
# Original and derived source layers will be given an appropriate prefix prior to output in the
# tier-specific rules sections.  Tier-specific parameters will likewise be given a prefix.
# Device subtypes will get an appropriate prefix as well.  Note that the wafer pin will not 
# necessarily be modeled correctly using this extraction deck.  The correct approach for device
# extraction is an important research topic for future study.
#  
# The INCLUDE_FILE directive is used in the preparation of the layers list prior to prefixing.  
# This is for cases where the file being processed is intended to be included after some other
# file has defined layer names.  In addition, the script will look in the rules file being
# processed for include statements and will follow them to find additional layers.
#
# Also note that if a layer is undefined (which would result in a parsing error in Calibre),
# that undefined layer will not get a prefix during replication.
#
# Lines that begin with the following keywords are commented out:
# TITLE, DRC, UNITS
#
# Known bugs and limitations:
#    Process definition blocks are not properly handled.  tier no. is not prepended to
#    directives.  
#
#    Check groups are not properly handled
#
#    Layer gathering is sensitive to case, cannot handle case mismatch
#
#    Layer rename misses a small number of instances (<10 for the big master drc deck)
#
#    Some #DEFINE statements need to be before certain includes.  This is not
#    properly handled 
#



require "pwd.pl";
&initpwd();

require "ctime.pl";


# first parse the invocation command
$infile="";
$config="";
$outpath="";
for ($i=0;$i<=5;$i+=2) {
    $_=$ARGV[$i];
    if (/^-infile$/) {
	$infile=$ARGV[$i+1];
    }
    elsif (/^-config$/) {
	$config=$ARGV[$i+1];
    }
    elsif (/^-outpath$/) {
	$outpath=$ARGV[$i+1];
    }
    else {
	die "usage:  convert_to_3d_rules.pl -infile input_file -config config_file -outpath output_dir\n";
    }
}

#read the config file
open(CONFIG,$config) || die "can't open config file\n";

#init to defaults
@global_params=();
$tier_offset=0;
$valid_range_max=-1;
$number_of_tiers=0;
@include_files=();
$create_ns_layers=1;

#flags to indicate that non-repeatable lines have been read
$flag_tier_offset=0;
$flag_valid_range_max=0;
$flag_number_of_tiers=0;
$flag_create_ns_layers=0;

while(<CONFIG>) {
    if (/^GLOBAL_PARAM /) {
	s/\n/ /;
	local $word;
	local @words=split;
	foreach $word (@words) {
	    ($word ne GLOBAL_PARAM) && push(@global_params,$word);
	}
    }
    elsif (/^INCLUDE_FILE /) {
	s/\n/ /;
	local $word;
	local @words=split;
	foreach $word (@words) {
	    ($word ne INCLUDE_FILE) && push(@include_files,$word);
	}
    }
    elsif (/^TIER_OFFSET /) {
	s/\n/ /;
	local $word;
	local @words=split;
	($flag_tier_offset) && die "Error:  redefinition of TIER_OFFSET not allowed\n";
	$tier_offset=$words[1];
	$flag_tier_offset=1;
    }
    elsif (/^VALID_RANGE_MAX /) {
	s/\n/ /;
	local $word;
	local @words=split;
	($flag_valid_range_max) && die "Error:  redefinition of VALID_RANGE_MAX not allowed\n";
	$valid_range_max=$words[1];
	$flag_valid_range_max=1;
    }
    elsif (/^NUMBER_OF_TIERS /) {
	s/\n/ /;
	local $word;
	local @words=split;
	($flag_number_of_tiers) && die "Error:  redefinition of NUMBER_OF_TIERS not allowed\n";
	$number_of_tiers=$words[1];
	$flag_number_of_tiers=1;
    }
    elsif (/^CREATE_NONSPECIFIC_LAYERS /) {
	s/\n/ /;
	local $word;
	local @words=split;
	($flag_create_ns_layers) && die "Error:  redefinition of CREATE_NONSPECIFIC_LAYERS not allowed\n";
	$create_ns_layers=$words[1];
	$flag_create_ns_layers=1;
    }
    elsif (/^REM/) {
    }
    elsif (/^\s*$/) {
    }
    else {
	s/\n/ /;
	local @words=split;
	local $badkeyword=$words[0];
	die "Error:  Invalid specification in config file: $badkeyword\n";
    }
}

# check for bad values
$flag_tier_offset || die "error:  must define TIER_OFFSET\n";
$flag_number_of_tiers || die "error:  must define NUMBER_OF_TIERS\n";
($tier_offset != int($tier_offset)) && die "error:  TIER_OFFSET must be an integer\n";
($valid_range_max != int($valid_range_max)) && die "error:  VALID_RANGE_MAX must be an integer\n";
($number_of_tiers != int($number_of_tiers)) && die "error:  NUMBER_OF_TIERS must be an integer\n";
($flag_valid_range_max) || ($valid_range_max=($tier_offset-1));
($tier_offset > $valid_range_max) || die "error:  TIER_OFFSET must be greater than VALID_RANGE_MAX\n";

close(CONFIG);

#determine if the input file is a master file
$_=$infile;
s/.*\/(.*)$/$1/;
$master_top=(/master_top/);

#select_output_file
# $filename_prefix will be prepended to ensure that we don't confuse the output with the input.
$filename_prefix="3D_";
$outfile=">".$outpath."/".$filename_prefix.$_;
$outfilepart=$_;

#process master_top file
if ($master_top) {

    open(INFILE, $infile) || die "can't open infile\n";
    open(OUTFILE, $outfile) || die "can't open outfile\n";

    # Create header section
    print OUTFILE "// ",$filename_prefix,$outfilepart,"\n";
    print OUTFILE "// ",&ctime(time),"\n";
    print OUTFILE "//\n";
    print OUTFILE "// This file was automatically generated based on the following input file:\n";
    print OUTFILE "// ",$infile,"\n//\n";
    print OUTFILE "// The following control settings were used in the automatic conversion:\n";
    print OUTFILE "//\n// NUMBER_OF_TIERS = ",$number_of_tiers,"\n";
    print OUTFILE "// TIER_OFFSET = ",$tier_offset,"\n";
    print OUTFILE "// VALID_RANGE_MAX = ",$valid_range_max,"\n";
    print OUTFILE "// CREATE_NONSPECIFIC_LAYERS = ",($create_ns_layers==1),"\n";
    print OUTFILE "//\n// Global parameter names reserved using GLOBAL_PARAM:\n";
    foreach $param (@global_params) {
	print OUTFILE "//\t",$param,"\n";
    }
    print OUTFILE "//\n// Included files specified using INCLUDE_FILE:\n";
    foreach $inc_file (@include_files) {
	print OUTFILE "//\t",$inc_file,"\n";
    }
    print OUTFILE "//\n// Based on the file name, this rules file has been determined to be a \"master_top\" file.\n";
    print OUTFILE "// It will be processed using the corresponding method.  Certain organizational conventions\n";
    print OUTFILE "// have been assumed.\n";
    print OUTFILE "//\n//\n/////////////////////////////////////////\n// THE CONVERTED RULE DECK BEGINS HERE //\n";
    print OUTFILE "/////////////////////////////////////////\n";

    # This section recognizes the following constructs:
    #     Block Comments ("/* */")
    #     Line and in-line comments ("//")
    #     #DEFINE preprocessor directives
    #     #IFDEF blocks for GLOBAL PARAMETERS
    #          (If the #IFDEF block looks at a local parameter, then an exception will be raised)
    #     LAYER statements
    #
    #     Other statements will be passed unchanged
    #
    #     It is assumed that this file performs no operations on layers.  If such operations exist,
    #     they will be passed to the output file without conversion for multiple tiers.

    # set comment block start flag to zero
    $flag_comment=0;
    # set process block flag to zero
    $process_block=0;

    while (<INFILE>) {
	if (/^\s*\/\//) {
	    # If the entire line is commented out, just print it
	    print OUTFILE;
	}
	elsif (/\*\//) {
	    # in-line block comments are not supported
	    (/\*\/\s*\S+$/) && die "error:  Non-white space characters found on same line after *\/.  This is not supported.\n";
	    ($flag_comment==0) && die "error:  End of block comment without beginning of block comment\n";
	    # flag end of block comment
	    $flag_comment=0;
	    print OUTFILE;
	}
	elsif (/\/\*/) {
	    # in-line block comments are not supported
	    (/^\S+\s*\/\*/) && die "error:  Non-white space characters found on same line before \/*.  This is not supported.\n"; 
	    ($flag_comment==1) && die "error:  nested block comment not supported\n";
	    # flag beginning of block comment
	    $flag_comment=1;
	    print OUTFILE;
	}
	elsif ($flag_comment) {
	    # if we are in the middle of a block comment, just print the line
	    print OUTFILE;
	}
	elsif ($process_block > 0) {
	    (/^\s*\#ENDIF/) && $process_block--;
	    print OUTFILE;
	}
	elsif (/^\s*\#IFDEF\s+USE_PROCESS/) {
	    #entering a process block
	    #write out without change, but keep track of nested directives
	    $process_block++;
	    s/\n/ /;
	    print OUTFILE $_,"\t//PROCESS DEFINITION BLOCK START\n";
	}
	elsif (/^\s*\#IFDEF/) {
	    ($process_block>0) && $process_block++;
	    local @words=split;
	    # if it is global, just write it, else replicate it
	    local @matches=grep(/^$words[1]$/,@global_params);
	    if ($#matches >= 0) {
		print OUTFILE;
	    }
	    else {
		die "error:  Creation of tier-specific \#IFDEF sections not supported for master top file.\n";
	    }
	}
	elsif (/^\s*\#DEFINE/) {
	    local @words=split;
	    # if it is global, just write it, else replicate it
	    local @matches=grep(/^$words[1]$/,@global_params);
	    if ($#matches >= 0) {
		print OUTFILE;
	    }
	    else {
		local $i;
		local $line=$_;
		for ($i=1;$i<=$number_of_tiers;$i++) {
		    $_=$line;
		    s/($words[1])/T$i\_$1/;
		    s/\n/ /;
		    print OUTFILE $_,"\t//tier ",$i,"\n";
		}
	    }
	}
	elsif (/^\s*LAYER/) {
	    local @words=split;
	    if ($words[2]<$valid_range_max) {
		local $i;
		local @modwords;
		for ($i=1;$i<=$number_of_tiers;$i++) {
		    @modwords=@words;
		    $modwords[2]+=($tier_offset*$i);
		    $modwords[2].="\t";
		    $modwords[1]="T".$i."_".$words[1]."\t\t";
		    $modwords[0].="\t";
		    $_=join(' ',@modwords);
		    print OUTFILE $_,"\t\t// Tier ",$i,"\n";
		}
		if ($create_ns_layers) {
		    @modwords=@words;
		    $modwords[2].="\t";
		    $modwords[1]="NS_".$words[1]."\t\t";
		    $modwords[0].="\t";
		    $_=join(' ',@modwords);
		    print OUTFILE $_,"\t\t// Non-specific Tier (phony layer) \n\n";
		}
	    }
	    else {
		print OUTFILE "// (OUT OF RANGE) -> ",$_;
	    }
	}
	else {
	    #if we haven't matched anything, then just print the line
	    print OUTFILE;
	}
    }
    
    # Don't allow unterminated comment block
    ($flag_comment==0) || (die "error:  unterminated block comment\n");

    close(INFILE);
    close(OUTFILE);
}
else {
    # for all other rules files
    
    open(INFILE, $infile) || die "can't open infile\n";
    open(OUTFILE, $outfile) || die "can't open outfile\n";

    # gather up the layers lists
    @original_layers=();
    @derived_lavers=();
    &get_layers($infile);

    foreach $file (@include_files) {
	&get_layers($file);
    }

    # begin preparing the output file
    print OUTFILE "//////////////////////////////////////////////\n";
    print OUTFILE "// MITLL 3D Calibre Rule Decks               \n";
    print OUTFILE "// Automatically generated from 2D rules files\n";
    print OUTFILE "// using convert_to_3d_rules utility          \n";
    print OUTFILE "// Automatic conversion date:  ",&ctime(time),"\n";
    print OUTFILE "// \n";
    print OUTFILE "// Please read carefully as further manual \n";
    print OUTFILE "// rules file modifications may be necessary.\n";
    print OUTFILE "// \n";
    print OUTFILE "// The following in-situ global \#DEFINE \n";
    print OUTFILE "// statements were found in the source rule\n"; 
    print OUTFILE "// deck:\n";
    print OUTFILE "// \n";
    
    &report_glob_def($infile);

    print OUTFILE "// \n";
    print OUTFILE "// The following include files may need to be \n";
    print OUTFILE "// modified so the path and filename are correct. \n";
    print OUTFILE "// \n";

    &copy_includes($infile);

    print OUTFILE "// \n";
    print OUTFILE "// Certain other SVRF commands should only appear\n";
    print OUTFILE "// once in the file.  It is recommended that the \n";
    print OUTFILE "// 3D deck be searched for the string \"(DISABLED) ->\" \n";
    print OUTFILE "// so that commands that were disabled since they are\n";
    print OUTFILE "// not local to a tier can be reenabled in a way that\n";
    print OUTFILE "// is appropriate to the whole stack.\n";

    print OUTFILE "// \n";
    print OUTFILE "// The following macro definitions apply to \n";
    print OUTFILE "// all tiers. \n";
    print OUTFILE "// \n";

    &copy_macros($infile);
	
    print OUTFILE "// Next, the source rules file is modified and\n";
    print OUTFILE "// instantiated for each tier.\n";
    print OUTFILE "// Layer names, non-global parameters, and \n";
    print OUTFILE "// device subtypes are given modified names.\n";
    
    &create_3d_rules($infile);

}

sub get_layers {
    # read a file and add original and derived layers to layer table
    #
    # Keep track of read layers in a running log file
    local $state="normal";
    # $state is "normal" when not inside a check, macro, etc.
    # block comments should also be excluded, but that has not been implemented
    local $layerslogfile=">>".$outpath."/layers_read.log";
    open(LAYERSLOGFILE, $layerslogfile) || die "can't open layers log file\n";
    
    local $inputfilename=$_[0];
    print LAYERSLOGFILE "\n\nReading input rules file:  ",$inputfilename,"\n";
    # If the input file name starts with a unix soft path, replace with a hard path
    # soft path must be at the beginning of the string
    if (/\$/) {
	local $_=$inputfilename;
	split /\//;
	$_=$_[0];
	s/\s*\$//;
	local $envpath = $ENV{$_};
	print LAYERSLOGFILE "foofoo:  ",$_,"\tfoo:  ",$envpath,"\n";
	$inputfilename =~ s/(\$.+)\//$envpath\//;
	print LAYERSLOGFILE "\n\nActual path: ",$inputfilename,"\n";
	print LAYERSLOGFILE "foo:  ",$ENV{"MGC_DESIGN_KIT"};
    }

    print LAYERSLOGFILE "\n\nReading input rules file:  ",$inputfilename,"\n";
    print LAYERSLOGFILE "Current time:  ",&ctime(time),"\n";

    open(my $layersreadfilehandle, $inputfilename) || die "can't open layers input file\n";
    $parencount=0;
    while(<$layersreadfilehandle>) {
	# 
	if ($parencount==0) {
	    $state="normal";
	}
	# The following should remove all comments from the line
	# there are some redundant checks for comments elsewhere
	# because those sections were written first
	s/\/\/.*$//;
	if ((/^\s*include/) || (/^\s*INCLUDE/)) {
	    local @words=split;
	    $_=@words[1];
	    s/\"//g;
	    print LAYERSLOGFILE "Found include statement -- need to include:  ",$_,"\n";
	    close LAYERSLOGFILE;
	    &get_layers($_);
	    open(LAYERSLOGFILE, $layerslogfile) || die "can't re-open layers log file\n";
	}
	elsif ($state eq "normal") {
#	    if (/^s*DMACRO/) {
#  move CMACRO, VARIABLE to top like DMACRO, Mar 2009, xchen
      if ((/^s*DMACRO/) || (/^s*CMACRO/) || (/^s*VARIABLE/)) {
		#don't want to capture local layers in macro blocks
		$state="macro";
		$parencount = $parencount + ($_ =~ tr/\{/\{/) - ($_ =~ tr/\}/\}/);
		#$parencount="1";  #this doesn't catch double curly brackets on 1st line
		#if ((/.*\}/) && (!(/.*\/\/.*\}/))) {
		#    $state="normal";
		#}
	    }
	    elsif (/^s*DEVICE/) {
		#don't want to capture parmaeter extractions in a device block
		$state="device";
		if ((/.*\]/) && (!(/.*\/\/.*\]/))) {
		    $state="normal";
		}
	    }
	    elsif (/^.+\{/) {
		#don't want to capture local layers or description contents in check blocks
		$state="check";
		$parencount = $parencount + ($_ =~ tr/\{/\{/) - ($_ =~ tr/\}/\}/);
		#if ((/.*\}/) && (!(/.*\/\/.*\}/))) {
		#    $state="normal";
		#}
	    }
	    elsif ((/^\s*layer/) || (/^\s*LAYER/)) {
		local @words=split;
		print LAYERSLOGFILE "Found original layer ",$words[1]," on GDS layer ",$words[2],"\n";
		$words[1] =~ s/\s+//g;
		push(@original_layers,$words[1]);
	    }
	    elsif (/^.+\s*=\s*.+$/) {
		if (!(/^\s*\/\//)) {
		    #ignore commented out operations
		    local @words=split/=/;
		    print LAYERSLOGFILE "Found derived layer ",$words[0]," defined by ",$_,"\n";
		    $words[0] =~ s/\s+//g;
		    push(@derived_layers,$words[0]);
		}
	    }
	}
	elsif ($state eq "macro") {
	    #curly bracket counting is not as robust as it should be
	    $parencount = $parencount + ($_ =~ tr/\{/\{/) - ($_ =~ tr/\}/\}/);
	    #if ($parencount==1 && ((/.*\}/) && (!(/.*\/\/.*\}/)))) {
	    #   $state="normal";
	    #	$parencount=0;
	    #}
	}
	elsif ($state eq "check") {
	    $parencount = $parencount + ($_ =~ tr/\{/\{/) - ($_ =~ tr/\}/\}/);
	    #if ((/.*\}/) && (!(/.*\/\/.*\}/))) {
	    #	$state="normal";
	    #}
	}
	elsif ($state eq "device") {
	    if ((/.*\]/) && (!(/.*\/\/.*\]/))) {
		$state="normal";
	    }
	}
    }

    print LAYERSLOGFILE "Done processing input rules file:  ",$inputfilename,"\n\n";
    print LAYERSLOGFILE "\n\nSummary of found layers:\n\nOriginal Layers:\n\n";
    print LAYERSLOGFILE join("\n",@original_layers),"\n\nDerived Layers:\n\n";
    print LAYERSLOGFILE join("\n",@derived_layers),"\n\n";
    close $layersreadfilehandle;
    close LAYERSLOGFILE;
    
}

sub report_glob_def {
    local $inputfilename=$_[0];
    local $linenumber=1;
    open (my $glob_infile, $inputfilename);
    while (<$glob_infile>) {
	if (/^\s*\#DEFINE/) {
	    local $line=$_;
	    local @words=split;
	    #if it is global, print to the report, otherwise ignore it
	    local @matches=grep(/^$words[1]$/,@global_params);
	    if ($#matches >= 0) {
		print OUTFILE "// line ",$linenumber,"\t--> ",$line,"\n";
	    }
	}
	$linenumber++;
    }
    print OUTFILE "//\n// This concludes the list of global \#DEFINE warnings.\n\n\n";
    close $glob_infile;
}

sub copy_macros {
    local $inputfilename=$_[0];
    open (my $macro_infile, $inputfilename);
    $parencount=0;
    while (<$macro_infile>) {
	s/\n/ /;
	if (/\/\//) {
	    $comment=$_;
	    $comment=~s/^.*\/\///;
	    $comment=" // ".$comment;
	}
	else {
	    $comment="";
	}
	
	$line=$_;
	$line =~ s/\/\/.*$//;

#	if ((/^s*DMACRO/) || $parencount>0) {
# Mar 2009, xchen	
	if ((/^s*DMACRO/) || (/^s*CMACRO/) || (/^s*VARIABLE/) || $parencount>0) {
	    $parencount = $parencount + ($line =~ tr/\{/\{/) - ($line =~ tr/\}/\}/);
	    print OUTFILE $line,$comment,"\n";
	}
    }
    print OUTFILE "//\n// This concludes the copied macros section.\n\n\n";
    close $macro_infile;
}

sub copy_includes {
    local $inputfilename=$_[0];
    open (my $includes_infile, $inputfilename);
    while (<$includes_infile>) {
	s/\n/ /;
	s/\/\/.*//;
	if (/^\s*[Ii][Nn][Cc][Ll][Uu][Dd][Ee]/) {
	    print OUTFILE "// ";
# remove the 2D layer definition inclusion from 3D rule file, xchen
	    print OUTFILE $_," // <- May not be correct\n"
	}
    }
    print OUTFILE "//\n// This concludes the copied includes section.\n\n\n";
    close $includes_infile;
}



sub create_3d_rules {
    local $inputfilename=$_[0];    
    
    # Create header section
    print OUTFILE "//\n//\n";
    print OUTFILE "// Job info:\n";
    print OUTFILE "// ",$filename_prefix,$outfilepart,"\n";
    print OUTFILE "// ",&ctime(time),"\n";
    print OUTFILE "//\n";
    print OUTFILE "// This section was automatically generated based on the following input file:\n";
    print OUTFILE "// ",$inputfilename,"\n//\n";
    print OUTFILE "// The following control settings were used in the automatic conversion:\n";
    print OUTFILE "//\n// NUMBER_OF_TIERS = ",$number_of_tiers,"\n";
    print OUTFILE "// TIER_OFFSET = ",$tier_offset,"\n";
    print OUTFILE "// VALID_RANGE_MAX = ",$valid_range_max,"\n";
    print OUTFILE "// CREATE_NONSPECIFIC_LAYERS = ",($create_ns_layers==1),"\n";
    print OUTFILE "//\n// Global parameter names reserved using GLOBAL_PARAM:\n";
    foreach $param (@global_params) {
	print OUTFILE "//\t",$param,"\n";
    }
    print OUTFILE "//\n// Included files specified using INCLUDE_FILE:\n";
    foreach $inc_file (@include_files) {
	print OUTFILE "//\t",$inc_file,"\n";
    }
    print OUTFILE "/////////////////////////////////////////\n";

    # This section recognizes the following constructs:
    #     Block Comments ("/* */")
    #     Line and in-line comments ("//")
    #     #DEFINE preprocessor directives
    #     #IFDEF blocks for GLOBAL PARAMETERS
    #          (If the #IFDEF block looks at a local parameter, then an exception will be raised)
    #     LAYER statements
    #
    #     Other statements will be passed unchanged
    #
    #     It is assumed that this file performs no operations on layers.  If such operations exist,
    #     they will be passed to the output file without conversion for multiple tiers.

    # set comment block start flag to zero
    $flag_comment=0;
    # set process block flag to zero
    $process_block=0;

    local $currenttier;
    for ($currenttier=1;$currenttier<=$number_of_tiers;$currenttier++) {
	$prefix="T".$currenttier."_";
	print OUTFILE "\n\n/////////////////////////////////////\n";
	print OUTFILE "//                                 //\n";
	print OUTFILE "// Rules for tier ",$currenttier,"                //\n";
	print OUTFILE "//                                 //\n";
	print OUTFILE "/////////////////////////////////////\n\n";
	open (my $infile, $inputfilename);
	while (<$infile>) {
	    if (/^\s*\/\//) {
		# If the entire line is commented out, just print it
		print OUTFILE;
	    }
	    elsif (/^\s*\@/) {
		# If it is a check description, just add the tier no.
		s/\@/\@TIER$currenttier /;
		print OUTFILE;
	    }
	    elsif (/\*\//) {
		# in-line block comments are not supported
		(/\*\/\s*\S+$/) && die "error:  Non-white space characters found on same line after *\/.  This is not supported.\n";
		($flag_comment==0) && die "error:  End of block comment without beginning of block comment\n";
		# flag end of block comment
		$flag_comment=0;
		print OUTFILE;
	    }
	    elsif (/\/\*/) {
		# in-line block comments are not supported
		(/^\S+\s*\/\*/) && die "error:  Non-white space characters found on same line before \/*.  This is not supported.\n"; 
		($flag_comment==1) && die "error:  nested block comment not supported\n";
		# flag beginning of block comment
		$flag_comment=1;
		print OUTFILE;
	    }
	    elsif ($flag_comment) {
		# if we are in the middle of a block comment, just print the line
		print OUTFILE;
	    }
	    elsif (/^\s*\#IFDEF/) {
		local @words=split;
		# if it is global, just write it, else replicate it
		local @matches=grep(/^$words[1]$/,@global_params);
		if ($#matches >= 0) {
		    print OUTFILE;
		}
		else {
		    s/(\#IFDEF )s*(S*)/$1$prefix$2/;
		    print OUTFILE;
		}
	    }
	    elsif (/^\s*\#DEFINE/) {
		local @words=split;
		# if it is global, just write it, else replicate it
		local @matches=grep(/^$words[1]$/,@global_params);
		if ($#matches >= 0) {
		    print OUTFILE;
		}
		else {
		    s/(\#DEFINE )s*(S*)/$1$prefix$2/;
		    print OUTFILE;
		}
	    }
	    elsif (/^\s*LAYER/) {
		local @words=split;
		if ($words[2]<$valid_range_max) {
		    local @modwords;
		    @modwords=@words;
		    $modwords[2]+=($tier_offset*$currenttier);
		    $modwords[2].="\t";
		    $modwords[1]="T".$currenttier."_".$words[1]."\t\t";
		    $modwords[0].="\t";
		    $_=join(' ',@modwords);
		    print OUTFILE $_,"\t\t// Tier ",$currenttier,"\n";
		}
		else {
		    print OUTFILE "// (OUT OF RANGE) -> ",$_;
		}
	    }
#	    elsif ((/^\s*DMACRO/) || ($dmacroparens > 0)) {
# Mar 2009, xchen
	    elsif ((/^s*DMACRO/) || (/^s*CMACRO/) || (/^s*VARIABLE/) || ($dmacroparens > 0)) {
		$dmacroparens = $dmacroparens + ($_ =~ tr/\{/\{/) - ($_ =~ tr/\}/\}/);
		print OUTFILE "// (MOVED TO TOP) -> ",$_;
	    }
	    elsif (/^\s*[Ii][Nn][Cc][Ll][Uu][Dd][Ee]/) {
		print OUTFILE "// (MOVED TO TOP) -> ",$_;
	    }
	    elsif ((/^\s*[Tt][Ii][Tt][Ll][Ee]/) ||
		   (/^\s*[Dd][Rr][Cc]\s+/) ||
		   (/^\s*[Ll][Vv][Ss]\s+/) ||
		   (/^\s*[Uu][Nn][Ii][Tt]/)) {
		print OUTFILE "// (DISABLED) -> ",$_;
	    }
	    else {
		#if we haven't matched anything, separate line and comment and add prefix to
		#layer names.  If this is a device, modify type/subtype to include tier info
		if (/^\s*DEVICE/) {
		    # if sybtype exists, append t# to subtype, else to type
		    local $suffix="_t".$currenttier;
		    if (/^\s*DEVICE\s+\S+\s+\(/) {
			s/(\s*DEVICE\s+\S+\s*\(\S+)(\))/$1$suffix$2/;
		    }
		    else {
			s/^(\s*DEVICE\s+\S+)(\s+)/$1$suffix$2/;
		    }   
		}
		elsif (/^.+\{/) {
		    #can't have duplicate check IDs
		    local $append="TIER".$currenttier."_";
		    s/^(.+\{)/$append$1/;
		}
		s/\n/ /;
		if (/\/\//) {
		    $comment=$_;
		    $comment=~s/^.*\/\///;
		    $comment=" // ".$comment;
		}
		else {
		    $comment="";
		}
	
		$line=$_;
		$line =~ s/\/\/.*$//;	
		
		#replace original layer names
		foreach $layer (@original_layers) {
		    $line =~ s/^($layer[\,\<\>\(\)=\s+])/$prefix$1/;
		    $line =~ s/([\,\<\>\(\)=\s+])($layer[\,\<\>\(\)=\s+])/$1$prefix$2/g;
		}
		foreach $layer (@derived_layers) {
		    $line =~ s/^($layer[\,\<\>\(\)=\s+])/$prefix$1/;
		    $line =~ s/([\,\<\>\(\)=\s+])($layer[\,\<\>\(\)=\s+])/$1$prefix$2/g;
		}
		print OUTFILE $line,$comment,"\n";
	    }
	}
	close $infile;
	print OUTFILE "///////////////////////////////\n";
	print OUTFILE "//   END OF TIER             //\n";
	print OUTFILE "///////////////////////////////\n\n\n";
    }
    
    # Don't allow unterminated comment block
    ($flag_comment==0) || (die "error:  unterminated block comment\n");
}
