#######################################################################
# cal.py
#######################################################################
"""
   calculate delay for memory array
"""

# PyCell code
from cni.dlo import (Box, DloGen, Instance, ParamArray, Point, Rect, Term, Tech, Layer, TermType, Pin, PropSet, InstanceArray, PhysicalComponent, Text, Orientation)
import math
from cell import (cols_cell, cell_width, width_ld, cols_sub, height_sa, rows_cell, cell_height, rows_sub, cols_top, rows_top, Ntier, tier_sub, output_bit, metal1_minwidth, metal2_minwidth)

# Rsqmh: The sheet resistance of H-tree wire
Rsqmh=0.25
# Rsqmp: The sheet resistance of pre-decode metal line
Rsqmp=0.25
# Rsqmg: The sheet resistance of global word-line
Rsqmg=0.25
# Rsqmwl: The sheet resistance of local word-line
Rsqmwl=0.38
# Rsqmbl: The sheet resistance of bit-line
Rsqmbl=0.25
# Rosa: Output resistance of sense-amplifier
Rosa=550
# Cg: the gate capacitance of NMOS per unit width (fF/um)
Cg=0.44
# Cmh: the unit area capacitance of H-tree wire (aF/um^2)
Cmh=41
# Cmp: the unit area capacitance of pre-decode line (aF/um^2)
Cpmp=41
# Cmg: the unit area capacitance of global word-line
Cmg=41
# Cmwl: the unit area capacitance of local word-line
Cmwl=77
# Cmbl: the unit area capacitance of bit-line
Cmbl=41
# Clsa: Load capacitance of sense-amplifier (fF)
Clsa=2.5
# wgdec: the width of global decoder input device (um)
wgdec=0.27
# wldec: the width of local decoder input device
wldec=0.27
# wsa: the width of sense-amp input device
wsa=0.27
# wcell_sw: the width of input MOS of bit-cell
wcell_sw=0.27
# Ndec: Number of logic stages in pre-decoder
Ndec=4
# tfo4: Extrinsic delay of the fanout-of-4 inverter (ps)
tfo4=5
# Pdec: Parasitic delay of pre-decoder
Pdec=0
# Vb: Voltage swing at bit-line
Vb=0.1
# Icell: Driving current of bit-cell (A)
Icell=0.0005
# z: proportionality constant
z=0.3
# Cj_cell: Junction capacitance of bit-cell output (fF)
Cj_cell=0.4
# Cj_mux: Junction capacitance of mux switch (fF)
Cj_mux=0.5
# rmux: turn-on resistance of mux
rmux=900
# CTSV: Capacitance of Through Silicon Via (fF)
CTSV=20
# RTSV: Resistance of Through Silicon Via (ohm)
RTSV=50
# r: Gama
r=1
# Rd: Resistance of a minimum-sized repeater (ohm)
Rd=550
# Cpd: Input capacitance of a minimum-sized repeater (fF)
Cpd=0.18

## delay calculation

# Wbank: width of a bank (w/o overhead) (um)
Wbank=(cols_cell*cell_width+width_ld)*cols_sub
# Lbank: height of a bank (w/o overhead) (um)
Lbank=(height_sa+rows_cell*cell_height)*rows_sub
# Lh: length of H-tree wire w/o repeater (um)
Lh=(cols_top/2-0.5)*Wbank+(rows_top/2-0.5)*Lbank
# Dhtree: Delay of I/O H-tree w/o repeaters (pS)
Dhtree=2*0.38*(Rsqmh*Lh)*(Cmh*Lh)/1000000
# Dhtree: Delay of I/O H-tree w/ repeaters (pS)
Dhtree2=(1.38+1.02*math.sqrt(1+r))*2*Lh*math.sqrt(Rd*Cpd*Rsqmp*Cmh/10**9)

# 3D Connection Delay
Dtsv=(Ntier-1)*(RTSV+Rd)*((Ntier-1)*0.5*CTSV+Cpd)/1000+Rd*((Ntier-1)*0.5*CTSV)/1000

# Ddec: Delay of decoder (pS)
Ddec=Ndec*tfo4+Pdec
# rp: resistance of pre-decode line (ohm)
rp=(height_sa+rows_cell*cell_height)*rows_sub*Rsqmp/2/metal2_minwidth
# cp: capacitance of pre-decode line (fF)
cp=(height_sa+rows_cell*cell_height)*rows_sub*metal2_minwidth*Cpmp/1000
# G: input capacitance of global decoder (fF)
G=wgdec*Cg*rows_sub
# rg: resistance of global word-line (ohm)
rg=max((cols_sub/2-1)*(cols_cell*cell_width+width_ld)*Rsqmg/metal2_minwidth,0)
# cg: capacitance of global word-line (fF)
cg=max((cols_sub-2)*(cols_cell*cell_width+width_ld)*metal2_minwidth*Cmg/1000,0)
# L: input capacitance of local decoder (fF)
L=wldec*Cg*cols_sub/2*tier_sub+cols_sub*CTSV*(tier_sub-1)/2
# rwl: resistance of local word-line (ohm)
rwl=cols_cell*cell_width*Rsqmwl/metal1_minwidth
# cwl: capacitance of local word-line (fF)
cwl=cols_cell*cell_width*metal1_minwidth*Cmwl/1000+2*cols_cell*wcell_sw*Cg
# Dwire: Delay of decoder wire and word-line (pS)
Dwire=rp*(cp/2+G)/1000+rg*(cg/2+L)/8000+rwl*cwl/8000
# cb: capacitance of bit-line (fF)
cb=rows_cell*cell_height*metal2_minwidth*Cmbl/1000
# Cb1: capacitance of bit-line and bit-cell output node (fF)
Cb1=cb+Cj_cell*rows_cell
# Cb2: capacitance of output mux (fF)
Cb2=Cj_mux*(cols_cell/output_bit+1)
# Cb3: csa, input capacitance of Sense-amplifier (fF)
Cb3=wsa*Cg
# Tr: local word-line rising time (pS)
Tr=(tfo4+rwl*cwl/8000)*1.4
# rbl: resistance of bit-line (fF)
rbl=rows_cell*cell_height*Rsqmbl/metal2_minwidth
# tRC: The time constant of bit-line RC network (pS)
tRC=rbl*Cb1*(Cb2+Cb3)/(Cb1+Cb2+Cb3)/1000+rmux*Cb3*(Cb1+Cb2)/(Cb1+Cb2+Cb3)/1000
# Dbl: Delay of bit-line (pS)
Dbl=(Cb1+Cb2+Cb3)*Vb/Icell/1000+z*Tr+tRC
# Tsa: Delay of sense-amplifier (pS)
Tsa=2*tfo4+3*Rosa*Clsa/1000

# Delay (ps) W/O H-tree delay
Delay1=Ddec+Dwire+Dbl+Tsa
# Total Delay (ps) with H-tree delay, no repeaters
TDelay=Delay1+Dhtree+Dtsv
# Total Delay (ps) with H-tree delay, With repeaters
TDelay2=Delay1+Dhtree2+Dtsv

